/*
 Wireless Gameboy/GBA Controller - Gamecube RX
 Version: 1.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 3/10/2018
 Last Modified: 27/11/2018
 
 Use any Gameboy / GBA which supports a Gameboy or GBA cartridge such as the Gameboy, GBC, GBP, GBA, GBA SP, 
 DS, DS Lite and GB Micro as a Wireless game controller for any device which supports USB HID Joysticks / Gamepad.
 
 This would allow you to play games on your Gamecube / Wii.
 
 
 Program the ATtiny441/841 to use a 16MHz crystal, enable BOD and set BOD level to 4.3V
 avrdude -p ATtiny441 -c usbasp -U lfuse:w:0xff:m -U hfuse:w:0xdc:m -U efuse:w:0xff:m
 
*/

// ATtiny441/841 Pin map
//
//                     +-\/-+
//               VCC  1|o   |14  GND
//    16MHz XTAL PB0  2|    |13  PA0 Input
//    16MHz XTAL PB1  3|    |12  PA1 Output
//         Reset PB3  4|    |11  PA2 LED
//               PB2  5|    |10  PA3 nRF24L01 CE
//  nRF24L01 CSN PA7  6|    |9   PA4 nRF24L01 SCK
//   nRF24L01 DI PA6  7|    |8   PA5 nRF24L01 DO
//                     +----+


#define F_CPU 16000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include "setup.c"
#include "mirf.c"


uint8_t modeFound = 0;
uint8_t kpData = 0;
uint8_t kpDataHigh = 0;

uint8_t volatile gcInputBuffer[10];
uint8_t kpOutputDpad = 0;
uint8_t kpOutputLRDownAndClick = 0;
uint8_t firstPacket = 1;

uint8_t kpLHeld = 0;
uint8_t kpLHeldCounter = 0;
uint8_t kpRHeld = 0;
uint8_t kpRHeldCounter = 0;

uint8_t gcReply1[3] = {0x09, 0x00, 0x20};
uint8_t gcReply2[10] = {0x00, 0x80, 0x7C, 0x78, 0x80, 0x79, 0x1A, 0x1B, 0x02, 0x02};
uint8_t gcControllerData[8] = {0x00, 0x80, 0x7D, 0x78, 0x80, 0x7A, 0x1A, 0x1B};


// While high, count bit time. Under 1.5uS low, higher than 2us high
uint8_t read_bit(void) {
	uint8_t bitTime = 0;
	
	while (!(PINA & (1<<DATA_IN))); // Wait until high
	//PORTA |= (1<<LED);
	asm volatile("nop");
	
	while (PINA & (1<<DATA_IN)) { // While high, count
		bitTime++;
	}
	//PORTA &= ~(1<<LED);
	asm volatile("nop");
	
	if (bitTime >= 8) { // High
		return 1;
	}
	else {
		return 0; // Low
	}
}

// Read data to the array until the count. Use the bypassHighLow if running this after interrupt
void read_bytes(uint8_t count) {
	while (!(PINA & (1<<DATA_IN))); // Wait until high
	while (PINA & (1<<DATA_IN)); // Wait until low
	
	for (uint8_t byte = 0; byte < count; byte++) {
		uint8_t dataIn = 0;
		
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
			if (read_bit()) {
				dataIn |= x;
			}
		}
		gcInputBuffer[byte] = dataIn;
	}
	
	while (!(PINA & (1<<DATA_IN))); // Wait until high
	_delay_us(1);
}

// Write data from the array until the count
void write_bytes(uint8_t* inputArray, uint8_t count) {
	for (uint8_t byte = 0; byte < count; byte++) {
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times
			if (inputArray[byte] & x) {
				PORTA |= (1<<DATA_OUT);
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				asm volatile("nop");
				PORTA &= ~(1<<DATA_OUT);
				_delay_us(3);
				PORTA |= (1<<DATA_OUT);
			}
			else {
				PORTA |= (1<<DATA_OUT);
				_delay_us(3);
				PORTA &= ~(1<<DATA_OUT);
				_delay_us(1);
				PORTA |= (1<<DATA_OUT);
			}
		}
	}
	PORTA |= (1<<DATA_OUT);
	_delay_us(2);
	PORTA &= ~(1<<DATA_OUT);
}


int main (void) {
	setup();
	
	check_for_new_settings(); // Check for new settings for 0.5 seconds
	mirf_config();
	
	// Check for GB/GBA mode
	check_for_gb_or_gba_mode();
	
	RX_POWERUP;
	mirf_flush_rx_tx(); // Flush TX/RX
	mirf_CE_hi; // Start listening
	
	_delay_ms(500); // Wait a little bit more for Gamecube to power up
	gcInputBuffer[0] = 0;
	
	while(1) {
		if (mirf_status() & (1<<RX_DR)) {
			PORTA |= (1<<LED);
			mirf_CE_lo; // Stop listening
			
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			
			if (modeFound == GBA_MODE) { // Load 2 bytes if in GBA mode
				kpDataHigh = spi_transfer(0);
				kpData = spi_transfer(0);
			}
			else {
				kpData = spi_transfer(0); // Read data
			}
			
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			mirf_flush_rx_tx(); // Flush TX/RX
			mirf_CE_hi; // Start listening
			PORTA &= ~(1<<LED);
			
			// The first packet received triggers special modes.
			if (firstPacket == 1) {
				firstPacket = 0;
				
				// Down held down, output the key presses to the D-Pad
				if (kpData & 0x20) {
					kpOutputDpad = 1;
				}
				// R held down, simulate the L/R button presses like in the GC controller, if they are held for more than 200ms
				// then also trigger the click switch too. Useful for Gamecube games that might use this.
				if (kpDataHigh & 0x02) { 
					kpOutputLRDownAndClick = 1;
				}
			}
			
			
			// Default values
			gcControllerData[0] = 0;
			gcControllerData[1] = 0x80;
			gcControllerData[2] = 0x7D;
			gcControllerData[3] = 0x78;
			gcControllerData[6] = 0x1A;
			gcControllerData[7] = 0x1B;
			kpLHeld = 0;
			kpRHeld = 0;
			
			// Check for key presses
			if (kpData & 0x01) { // A
				gcControllerData[0] |= 0x01;
			}
			if (kpData & 0x02) { // B
				gcControllerData[0] |= 0x02;
			}
			if (kpData & 0x04) { // Right
				if (kpOutputDpad == 1) {
					gcControllerData[1] |= 0x02; // D-Pad
				}
				else {
					gcControllerData[2] = 0xFF;
				}
			}
			if (kpData & 0x08) { // Left
				if (kpOutputDpad == 1) {
					gcControllerData[1] |= 0x01; // D-Pad
				}
				else {
					gcControllerData[2] = 0x00;
				}
			}
			if (kpData & 0x10) { // Up
				if (kpOutputDpad == 1) {
					gcControllerData[1] |= 0x08; // D-Pad
				}
				else {
					gcControllerData[3] = 0xFF;
				}
			}
			if (kpData & 0x20) { // Down
				if (kpOutputDpad == 1) {
					gcControllerData[1] |= 0x04; // D-Pad
				}
				else {
					gcControllerData[3] = 0x00;
				}
			}
			if (kpData & 0x40) { // Select
				gcControllerData[0] |= 0x08; // Y
			}
			if (kpData & 0x80) { // Start
				gcControllerData[0] |= 0x10;
			}
			if (kpDataHigh & 0x01) { // L (for GBA)
				if (kpOutputLRDownAndClick == 1) {
					kpLHeld = 1;
					kpLHeldCounter++;
					gcControllerData[6] = 0x78;
					
					if (kpLHeldCounter >= 250) {
						gcControllerData[1] |= 0x40;
						kpLHeldCounter = 250;
					}
				}
				else {
					gcControllerData[1] |= 0x40;
				}
			}
			if (kpDataHigh & 0x02) { // R (for GBA)
				if (kpOutputLRDownAndClick == 1) {
					kpRHeld = 1;
					kpRHeldCounter++;
					
					gcControllerData[7] = 0x78;
					if (kpRHeldCounter >= 250) {
						gcControllerData[1] |= 0x20;
						kpRHeldCounter = 250;
					}
				}
				else {
					gcControllerData[1] |= 0x20;
				}
			}
			
			// Reset L/R counter if not held down
			if (kpOutputLRDownAndClick == 1) {
				if (kpLHeld == 0) {
					kpLHeldCounter = 0;
				}
				if (kpRHeld == 0) {
					kpRHeldCounter = 0;
				}
			}
		}
		_delay_us(250);
		
		
		// At first boot and if we ever receive 0 as the first byte (e.g when GC reboots)
		if (gcInputBuffer[0] == 0) {
			GIMSK &= !(1<<PCIE0); // Turn off interrupts
			
			// Wait for initial request from Gamecube (0000 0000)
			// Transmit Gamecube controller reply 1
			read_bytes(1);
			while (gcInputBuffer[0] != 0) {
				_delay_ms(10);
				read_bytes(1);
			}
			write_bytes(gcReply1, 3);
			
			
			// Gamecube responds (0100 0001)
			// Transmit Gamecube controller reply 2
			read_bytes(1);
			write_bytes(gcReply2, 10);
			
			// Turn on pin interrupt
			_delay_us(50); // Wait a little
			PCMSK0 |= (1<<PCINT0);
			GIMSK |= (1<<PCIE0);
		}
	}
	
	return 0;
}

// We woke up because GC pulled the input low
ISR(PCINT0_vect) {
	if (!(PINA & (1<<DATA_IN))) {
		PORTA &= ~(1<<LED);
		// Read 3 bytes, assumed it's the Gamecube poll (0100 0000 0000 0011 0000 0010)
		for (uint8_t byte = 0; byte < 3; byte++) {
			uint8_t dataIn = 0;
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				if (read_bit()) {
					dataIn |= x;
				}
			}
			gcInputBuffer[byte] = dataIn;
		}
		while (!(PINA & (1<<DATA_IN))); // Wait until high
		_delay_us(1);
		
		// Write the controller data
		write_bytes(gcControllerData, 8);
		
		_delay_us(50); // Wait a little so we don't re-trigger the interrupt
	}
}
